(*  Title:      ISM/NSL.thy
    Id:         $Id: NSL.thy,v 1.31 2003/05/28 11:40:06 dvo Exp $
    Author:     David von Oheimb
    Copyright:  (C) 2002 Siemens AG; any non-commercial use is granted

*)

header {* Lowe's fix of the Needham-Schroeder public-key protocol *}

theory NSL
imports ISM_package DYSProtocols

begin

(*<* )
ML {* Unify.trace_bound  := 100; *}
ML {* Pretty.setmargin 130; quick_and_dirty:=true *}

( *>*)
constdefs 
 "Alice" :: agent
 "Alice \<equiv> Friend 0"

 "Bob" :: agent
 "Bob \<equiv> Friend 1"

 "Intruder" :: agent
 "Intruder \<equiv> Spy"

 "NGen" :: agent
 "NGen \<equiv> Server"

(*<*)
lemma [simp]: "Bob \<noteq> Alice"      by (simp add: Bob_def Alice_def)
lemma [simp]: "Alice \<noteq> Bob"      by (simp add: Bob_def Alice_def)
lemma [simp]: "Bob \<noteq> Intruder"   by (simp add: Bob_def Intruder_def)
lemma [simp]: "Intruder \<noteq> Bob"   by (simp add: Bob_def Intruder_def)
lemma [simp]: "Alice \<noteq> Intruder" by (simp add: Alice_def Intruder_def)
lemma [simp]: "Intruder \<noteq> Alice" by (simp add: Alice_def Intruder_def)
lemma [simp]: "NGen \<noteq> Alice"     by (simp add: NGen_def Alice_def)
lemma [simp]: "Alice \<noteq> NGen"     by (simp add: NGen_def Alice_def)
lemma [simp]: "NGen \<noteq> Bob"       by (simp add: NGen_def Bob_def)
lemma [simp]: "Bob \<noteq> NGen"       by (simp add: NGen_def Bob_def)
lemma [simp]: "NGen \<noteq> Intruder"  by (simp add: NGen_def Intruder_def)
lemma [simp]: "Intruder \<noteq> NGen"  by (simp add: NGen_def Intruder_def)

(*>*)
datatype A_control = Init | Wait | Fine
record A_data = 
  Apeer :: "agent"
  AnA   :: "nat"
  AnB   :: "nat"
(*<*)
translations "A_data" \<leftharpoondown> (type) "\<lparr>Apeer :: agent, AnA :: nat, AnB :: nat\<rparr>"

(*>*)
consts A0 :: A_data
types A_state = "A_control \<times> A_data"
(*<*)
translations "A_state" \<leftharpoondown> (type) "A_control \<times> A_data"
(*>*)

datatype B_control = Idle | Resp | Conn
record B_data = 
  Bpeer :: "agent"
  BnA   :: "nat"
  BnB   :: "nat"
(*<*)
translations "B_data" \<leftharpoondown> (type) "\<lparr>Bpeer :: agent, BnA :: nat, BnB :: nat\<rparr>"

(*>*)
consts B0 :: B_data
types B_state = "B_control \<times> B_data"
(*<*)
translations "B_state" \<leftharpoondown> (type) "B_control \<times> B_data"
(*>*)

types I_state = "msg set"
(*<*)
translations "I_state" \<leftharpoondown> (type) "msg set"
(*>*)

types N_state = "nat set"
(*<*)
translations "N_state" \<leftharpoondown> (type) "nat set"
(*>*)
consts N0 :: "nat set"

datatype state = AS A_state | IS I_state | BS B_state | NS N_state
consts
  the_AS :: "state \<Rightarrow> A_state"
  the_IS :: "state \<Rightarrow> I_state"
  the_BS :: "state \<Rightarrow> B_state"
  the_NS :: "state \<Rightarrow> N_state"
primrec "the_AS (AS x) = x"
primrec "the_IS (IS x) = x"
primrec "the_BS (BS x) = x"
primrec "the_NS (NS x) = x"

consts
  Alice_state    :: "(agent \<Rightarrow> state) \<Rightarrow> A_state"
  Bob_state      :: "(agent \<Rightarrow> state) \<Rightarrow> B_state"
  Intruder_known :: "(agent \<Rightarrow> state) \<Rightarrow> msg set"
  NGen_used      :: "(agent \<Rightarrow> state) \<Rightarrow> nat set"
defs
 Alice_state_def:    "Alice_state    s \<equiv> the_AS (s Alice)"
 Bob_state_def:      "Bob_state      s \<equiv> the_BS (s Bob)"
 Intruder_known_def: "Intruder_known s \<equiv> the_IS (s Intruder)"
 NGen_used_def:      "NGen_used      s \<equiv> the_NS (s NGen)"

(*<*)
lemma Alice_state_Alice [simp]: 
  "Alice_state (s(Alice:=AS x)) = x" by (simp add: Alice_state_def)
lemma Alice_state_Bob [simp]: 
  "Alice_state (s(Bob:=x)) = Alice_state s" by (simp add: Alice_state_def)
lemma Alice_state_Intruder [simp]: 
  "Alice_state (s(Intruder:=x)) = Alice_state s" by (simp add: Alice_state_def)
lemma Alice_state_NGen [simp]: 
  "Alice_state (s(NGen:=x)) = Alice_state s" by (simp add: Alice_state_def)

lemma Bob_state_Bob [simp]: 
  "Bob_state (s(Bob:=BS x)) = x" by (simp add: Bob_state_def)
lemma Bob_state_Alice [simp]: 
  "Bob_state (s(Alice:=x)) = Bob_state s" by (simp add: Bob_state_def)
lemma Bob_state_Intruder [simp]: 
  "Bob_state (s(Intruder:=x)) = Bob_state s" by (simp add: Bob_state_def)
lemma Bob_state_NGen [simp]: 
  "Bob_state (s(NGen:=x)) = Bob_state s" by (simp add: Bob_state_def)

lemma Intruder_known_Alice [simp]: 
  "Intruder_known (s(Alice:=x)) = Intruder_known s" 
by (simp add: Intruder_known_def)
lemma Intruder_known_Bob [simp]: 
  "Intruder_known (s(Bob:=x)) = Intruder_known s" 
by (simp add: Intruder_known_def)
lemma Intruder_known_Intruder [simp]: 
  "Intruder_known (s(Intruder:=IS known)) = known" 
by (simp add: Intruder_known_def)
lemma Intruder_known_NGen [simp]: 
  "Intruder_known (s(NGen:=x)) = Intruder_known s" 
by (simp add: Intruder_known_def)

lemma NGen_used_Alice [simp]: 
  "NGen_used (s(Alice:=x)) = NGen_used s" by (simp add: NGen_used_def)
lemma NGen_used_Bob [simp]: 
  "NGen_used (s(Bob:=x)) = NGen_used s" by (simp add: NGen_used_def)
lemma NGen_used_Intruder [simp]: 
  "NGen_used (s(Intruder:=x)) = NGen_used s" by (simp add: NGen_used_def)
lemma NGen_used_NGen [simp]: 
  "NGen_used (s(NGen:=NS used)) = used" by (simp add: NGen_used_def)
(*>*)
datatype channel = AI | IB | BI | IA | NA | NB


subsection "Alice"

ism Alice =
  ports channel
    inputs "{NA,IA}"
    outputs "{AI}"
  messages msg
  states   state
    control A_control init "Init"
    data    A_data    init "A0"
  transitions
  Req: Init \<rightarrow> Wait
    in  NA "[Nonce nA]"
    out AI "[Crypt (pubK B) \<lbrace>Nonce nA, Agent Alice\<rbrace>]"
    post Apeer := B, AnA := nA
  Ack: Wait \<rightarrow> Fine
    pre "nA = AnA s", "B = Apeer s"
    in  IA "[Crypt (pubK Alice) \<lbrace>Nonce nA, Nonce nB, Agent B\<rbrace>]"
    out AI "[Crypt (pubK B) (Nonce nB)]"
    post AnB := nB


subsection "Bob"

ism Bob =
  ports channel
    inputs "{NB,IB}"
    outputs "{BI}"
  messages msg
  states   state
    control B_control init "Idle"
    data    B_data    init "B0"
  transitions
  Resp: Idle \<rightarrow> Resp
    in  NB "[Nonce nB]", IB "[Crypt (pubK Bob) \<lbrace>Nonce nA,Agent A\<rbrace>]"
    out BI "[Crypt (pubK A) \<lbrace>Nonce nA, Nonce nB, Agent Bob\<rbrace>]"
    post Bpeer := A, BnA := nA, BnB := nB
  Ack': Resp \<rightarrow> Conn
    pre "nB = BnB s"
    in  IB "[Crypt (pubK Bob) (Nonce nB)]"


subsection "Intruder"

constdefs 
  I_inputs :: "channel set"
 "I_inputs \<equiv> {AI, BI}"
  I_outputs :: "channel set"
 "I_outputs \<equiv> {IA, IB}"
ism Intruder =
  ports channel
    inputs  "I_inputs"
    outputs "I_outputs"
  messages msg
  states   state
    data "msg set" init "initState Intruder" name "known"
  transitions
  Learn: 
    pre "ch \<in> I_inputs"
    in   ch "[m]"
    post "insert m known"
  Utter:
    pre "ch \<in> I_outputs", "m \<in> synth (analz known)"
    out  ch "[m]"

text {*\pagebreak*}
subsection "Nonce Generator"

constdefs 
  N_outputs :: "channel set"
 "N_outputs \<equiv> {NA, NB}"  (*### output also to Intruder? *)

ism NGen =
  ports channel
    inputs  "{}"
    outputs "N_outputs"
  messages msg
  states   state
    data "nat set" init "N0" name "used"
  transitions
  Cackle:
    pre "ch \<in> N_outputs", "n \<notin> used"
    out  ch "[Nonce n]"
    post "insert n used"


subsection "System"

types NSL_state = "agent \<Rightarrow> state"
      NSL_conf  = "(channel, msg, NSL_state) conf"
(*<*)
translations "NSL_state"  \<leftharpoondown> (type) "agent \<Rightarrow> state"
             "NSL_conf"   \<leftharpoondown> (type) "(channel, msg, NSL_state) conf"
(*>*)

constdefs 
  System ::"(agent, (unit, channel, msg, state, unit) ism) family"
 "System  \<equiv> (undefined(Alice:=Alice.ism,Intruder:=Intruder.ism,
                       Bob:=Bob.ism,NGen:=NGen.ism),{Alice,Intruder,Bob,NGen})"

  NSL_Runs :: "(NSL_conf list) set"
 "NSL_Runs \<equiv> runs (ISM_comp System)"

(*<*)
lemmas System_ism_defs = Alice.ism_def Intruder.ism_def Bob.ism_def NGen.ism_def ism.make_def
lemmas System_trans_elims = Alice.transs.cases Bob.transs.cases Intruder.transs.cases NGen.transs.cases

lemma all_inputs_System [simp]: "all_inputs System = {AI, IA, IB, BI, NA, NB}"
by (simp add: all_inputs_def System_def System_ism_defs I_inputs_def, blast)

lemma all_outputs_System [simp]: "all_outputs System = {AI, IA, IB, BI, NA, NB}"
by (simp add: all_outputs_def System_def System_ism_defs I_outputs_def N_outputs_def, blast)

lemma wf_isms_System: "wf_isms System"
(*<*)
by (auto intro!: wf_ismsI wf_ismI elim!: System_trans_elims
      simp add: System_def System_ism_defs fun_upd_apply split add: split_if_asm)
(*>*)
lemma wf_comp_System: "wf_comp System"  --{* unused *}
(*<*)
by (auto simp add: wf_comp_def System_def System_ism_defs 
                     I_inputs_def fun_upd_apply split add: split_if_asm)
(*>*)

lemma closed_System: "closed System"
(*<*)
by (simp add: closed_def)
(*>*)

lemma in_SystemD: "i \<in> snd System \<Longrightarrow> 
 i = Alice    \<and> fst System i = Alice.ism    \<or> 
 i = Bob      \<and> fst System i = Bob.ism      \<or> 
 i = Intruder \<and> fst System i = Intruder.ism \<or> 
 i = NGen     \<and> fst System i = NGen.ism"
by (auto del: conjI simp add: System_def fun_upd_apply)

lemma NSL_Runs_System_past_induct:
"\<lbrakk>(bs,s)#cs \<in> NSL_Runs;
  P \<currency> ((init \<circ> (undefined::(agent => (unit, channel, msg, state, unit) ism)))
                (Alice:=AS (Init, A0), Intruder:=IS (initState Intruder),
                 Bob:=BS (Idle, B0), NGen:=NS N0)) [];
 \<And>p1 p2 cs b c s s1 s2.
       \<lbrakk>(p1 .@. b, s(Alice:=s1))#cs \<in> NSL_Runs; 
        P (p1 .@. b) (s(Alice:=s1)) cs; ((p1, s1), c, p2, s2) \<in> Alice.transs\<rbrakk>
       \<Longrightarrow> P (b .@. p2) (s(Alice:=s2)) ((p1 .@. b, s(Alice:=s1)) # cs);
 \<And>p1 p2 cs b c s s1 s2.
       \<lbrakk>(p1 .@. b, s(Bob:=s1))#cs \<in> NSL_Runs; 
        P (p1 .@. b) (s(Bob:=s1)) cs; ((p1, s1), c, p2, s2) \<in> Bob.transs\<rbrakk>
       \<Longrightarrow> P (b .@. p2) (s(Bob:=s2)) ((p1 .@. b, s(Bob:=s1)) # cs);
 \<And>p1 p2 cs b c s s1 s2.
       \<lbrakk>(p1 .@. b, s(Intruder:=s1))#cs \<in> NSL_Runs; 
        P (p1 .@. b) (s(Intruder:=s1)) cs; ((p1, s1), c, p2, s2) \<in> Intruder.transs\<rbrakk>
       \<Longrightarrow> P (b .@. p2) (s(Intruder:=s2)) ((p1 .@. b, s(Intruder:=s1)) # cs);
 \<And>p1 p2 cs b c s s1 s2.
       \<lbrakk>(p1 .@. b, s(NGen:=s1))#cs \<in> NSL_Runs; 
        P (p1 .@. b) (s(NGen:=s1)) cs; ((p1, s1), c, p2, s2) \<in> NGen.transs\<rbrakk>
       \<Longrightarrow> P (b .@. p2) (s(NGen:=s2)) ((p1 .@. b, s(NGen:=s1)) # cs)
\<rbrakk> \<Longrightarrow> P bs s cs"
apply (unfold NSL_Runs_def, erule runs_closed_comp_past_induct, 
       rule_tac [5] refl)
apply    (rule closed_System)
apply   (rule wf_isms_System)
apply  (simp add: System_def System_ism_defs, fold K_Nil_def o_def, 
        simp del: o_apply o_def)
apply (drule in_SystemD)
apply safe
apply (simp_all add: System_ism_defs del: fappend_def2)
done

declare fappend_upd_conv [simp del]
declare upd_fappend_conv [simp del]


lemma Bob_state_Resp_eq_lemma: 
  "(b, s) # cs \<in> NSL_Runs \<Longrightarrow> Bob_state s = (Idle, sa) \<longrightarrow>
 (b', s') \<in> set cs \<longrightarrow> Bob_state s' \<noteq> (Resp, sa')"
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply auto
done
(*>*)

lemma Bob_state_Resp_eq [rule_format]: "(b ,s )#cs \<in> NSL_Runs \<Longrightarrow>
 Bob_state s = (Resp, bs ) \<longrightarrow> (b',s')\<in>set cs \<longrightarrow> 
 Bob_state s'= (Resp, bs') \<longrightarrow> bs = bs'"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: Bob_state_Resp_eq_lemma)
done

(*>*)
text {*\pagebreak*}
subsection "Distinctness (i.e., freshness) of nonces"


lemma NB_used [rule_format]:
  "(b, s) # cs \<in> NSL_Runs \<Longrightarrow> Nonce n \<in> set (b NB) \<longrightarrow> n \<in> NGen_used s"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto)
apply (auto simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)
lemma NA_used [rule_format]:
  "(b, s) # cs \<in> NSL_Runs \<Longrightarrow> Nonce n \<in> set (b NA) \<longrightarrow> n \<in> NGen_used s"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto)
apply (auto simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)
lemma NA_used_past [rule_format]: "(b, s) # cs \<in> NSL_Runs \<Longrightarrow> 
  Nonce n \<in> set (b' NA) \<longrightarrow> (b', s') \<in> set cs \<longrightarrow> n \<in> NGen_used s"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: NA_used)
done
(*>*)

lemma NB_disjoint_NA: "(b ,s )#cs \<in> NSL_Runs \<Longrightarrow>
 Nonce n \<in> set (b NB) \<longrightarrow> Nonce n \<notin> set (b NA)"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: NB_used NA_used 
            simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)
lemma NB_disjoint_NA_past: "(b ,s )#cs \<in> NSL_Runs \<Longrightarrow>
 Nonce n \<in> set (b NB) \<longrightarrow> (b',s')\<in>set cs \<longrightarrow> Nonce n \<notin> set (b' NA)"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: NB_disjoint_NA NA_used NA_used_past 
            simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)

lemma BnB_used [rule_format]: "(b ,s )#cs \<in> NSL_Runs \<Longrightarrow>
 Bob_state s = (Resp, bs) \<longrightarrow> BnB bs \<in> NGen_used s"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: NB_used 
            simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)

lemma BnB_not_in_NA: "(b ,s )#cs \<in> NSL_Runs \<Longrightarrow>
 Bob_state s = (Resp, bs) \<longrightarrow> Nonce (BnB bs) \<notin> set (b NA)"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: NB_disjoint_NA BnB_used 
            simp add: I_inputs_def I_outputs_def N_outputs_def)
done

(*>*)
lemma BnB_not_in_NA_past: "(b,s)#cs \<in> NSL_Runs \<Longrightarrow>
  Bob_state s = (Resp, bs) \<longrightarrow> (b',s')\<in>set cs \<longrightarrow> Nonce (BnB bs) \<notin> set (b' NA)"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (safe elim!: System_trans_elims del: impCE)
apply (auto dest: BnB_not_in_NA NB_disjoint_NA NB_disjoint_NA_past)
done

(*>*)

constdefs
  Alice_Wait_nA  :: "nonce \<Rightarrow> NSL_state \<Rightarrow> bool"
 "Alice_Wait_nA nB s \<equiv> \<exists>as. Alice_state s = (Wait,as) \<and> AnA as = nB"

lemma Alice_Wait_nA_NA: "cs \<in> NSL_Runs \<Longrightarrow>
 (\<forall>c\<in>set cs. Nonce nB \<notin> set (fst c NA)) \<longrightarrow> 
 (\<forall>c\<in>set cs. \<not>Alice_Wait_nA nB (snd c))"
(*<*)
apply (unfold NSL_Runs_def Alice_Wait_nA_def, 
       erule runs_closed_comp_all_imp_all_induct, simp_all)
apply    (rule closed_System)
apply   (rule wf_isms_System)
apply  (fold K_Nil_def o_def)
apply  (simp add: System_def System_ism_defs del: o_apply o_def)
apply (drule in_SystemD)
apply (safe)
apply (simp_all add: System_ism_defs)
apply (erule_tac [!] System_trans_elims)
apply (auto)
done
(*>*)

lemma Alice_Wait_nA_not_BnB [rule_format]: 
 "\<lbrakk>(b,s)#cs \<in> NSL_Runs; Bob_state s = (Resp, bs)\<rbrakk> \<Longrightarrow> 
  \<forall>c\<in>set ((b,s)#cs). \<not>Alice_Wait_nA (BnB bs) (snd c)"
(*<*)
apply (rule Alice_Wait_nA_NA [THEN mp])
apply  simp
apply (auto dest: BnB_not_in_NA BnB_not_in_NA_past)
done
(*>*)

subsection "A variant of Steve Schneider's rank function proof technique"

consts
  Alice_Fine_with  :: "agent \<Rightarrow> nonce \<Rightarrow> NSL_state \<Rightarrow> bool"
  Bob_Resp_only_to :: "agent          \<Rightarrow> NSL_state \<Rightarrow> bool"
  constrain        ::          "nonce \<Rightarrow> NSL_state \<Rightarrow> bool"
defs
  Alice_Fine_with_def: "Alice_Fine_with B nB s \<equiv>
                   \<exists>nA. Alice_state s = (Fine, \<lparr>Apeer = B, AnA = nA, AnB = nB\<rparr>)"

  Bob_Resp_only_to_def: 
 "Bob_Resp_only_to A s \<equiv> \<forall>bs. Bob_state s = (Resp, bs) \<longrightarrow> Bpeer bs = A"

  constrain_def: "constrain nB s \<equiv> 
  \<not>Alice_Wait_nA nB s \<and> \<not>Alice_Fine_with Bob nB s \<and> Bob_Resp_only_to Alice s"

text{*\pagebreak*}
text {* @{term "noleak nB"} describes 
   a superset of the messages transmitted by any agent (including the intruder)
   during runs of the constrained protocol, such that 
   these messages cannot lead to sending @{term "Crypt (pubK Bob) (Nonce nB)"}, 
   which in particular means that @{term nB} is not leaked to the intruder *}
consts 
  noleak :: "nonce \<Rightarrow> msg \<Rightarrow> bool"
primrec
 "noleak nB (Number  n) = True" --{* cannot occur in this protocol *}
 "noleak nB (Hash    m) = True" --{* cannot occur in this protocol *}
 "noleak nB (Agent   a) = True"
 "noleak nB (Nonce   n) = (n \<noteq> nB)"
 "noleak nB (Key     k) = (Key k \<in> initState Intruder)"
 "noleak nB (\<lbrace>m, n\<rbrace>   ) = (noleak nB m \<and> noleak nB n)"
 "noleak nB (Crypt k m) = (k = pubK Alice \<and> (\<exists>n. m = \<lbrace>n, Nonce nB, Agent Bob\<rbrace>)
                           \<or> noleak nB m)"

text {* Not explicitly used, just for validation of @{term noleak}: *}
lemma noleak_Crypt_nA_Alice: 
  "nA \<noteq> nB \<Longrightarrow> noleak nB (Crypt (pubK B) \<lbrace>Nonce nA, Agent Alice\<rbrace>)"
(*<*)
by simp
(*>*)
lemma noleak_Crypt_n_nB_Bob: 
  "noleak nB (Crypt (pubK Alice) \<lbrace>Nonce nA, Nonce nB, Agent Bob\<rbrace>)"
(*<*)
by simp
(*>*)
lemma leak_Crypt_Bob_nB: "\<not> noleak nB (Crypt (pubK Bob) (Nonce nB))"
(*<*)
by simp

lemma initState_Intruder [simp]: "initState Intruder = initState Spy"
by (simp add: Intruder_def)
(*>*)
(* Not explicitly used: *)
lemma init_noleak: "\<forall>m \<in> initState Intruder. noleak nB m"
(*<*)
by force

lemma analz_maintains_noleak [rule_format]: 
  "Alice \<notin> bad \<Longrightarrow> m \<in> analz known \<Longrightarrow> Ball known (noleak nB) \<longrightarrow> noleak nB m"
apply (erule analz.induct)
apply auto
done

lemma synth_maintains_noleak [rule_format]: 
  "m \<in> synth known \<Longrightarrow> 
   Ball known (noleak nB) \<Longrightarrow> noleak nB m"
apply (erule synth.induct)
apply auto
done

(*>*)
lemma synth_analz_maintains_noleak: 
"\<lbrakk>m \<in> synth (analz known); \<forall>m\<in>known. noleak nB m; Alice \<notin> bad\<rbrakk> \<Longrightarrow> noleak nB m"
(*<*)
apply (erule synth_maintains_noleak)
apply (erule (1) analz_maintains_noleak, fast)
done
(*>*)

constdefs
  noleak_all :: "nonce \<Rightarrow> NSL_conf \<Rightarrow> bool"
 "noleak_all nB c \<equiv> (\<forall>p. p \<notin> N_outputs \<longrightarrow> (\<forall>m \<in> set (fst c p). noleak nB m)) 
                  \<and> (\<forall>m \<in> Intruder_known (snd c). noleak nB m)"

(*<*)
lemma noleak_all_fappend [simp]:
  "noleak_all nB (a .@. b, s) = (noleak_all nB (a, s) \<and> noleak_all nB (b, s))"
by (auto simp add: noleak_all_def)

lemma noleak_all_upd_Alice [simp]: 
 "noleak_all nB (b, s(Alice:=y)) = noleak_all nB (b, s)"
by (simp_all add: noleak_all_def)
lemma noleak_all_upd_Bob [simp]: 
 "noleak_all nB (b, s(Bob:=y)) = noleak_all nB (b, s)"
by (simp_all add: noleak_all_def)

lemma noleak_all_1msg [simp]:
"\<lbrakk>p \<noteq>  NA; p \<noteq>  NB\<rbrakk> \<Longrightarrow> noleak_all nB (\<currency>(p := [m]), s) = 
 (noleak nB m \<and> (\<forall>m \<in>  Intruder_known s. noleak nB m))"
by (auto simp add: noleak_all_def N_outputs_def fun_upd_apply)

lemma noleak_all_add1msg_head [simp]: "\<lbrakk>p \<noteq>  NA; p \<noteq>  NB\<rbrakk> \<Longrightarrow>  
noleak_all nB (b(p := m#b p), s) = (noleak_all nB (b, s) \<and> noleak nB m)"
apply (simp add: noleak_all_def N_outputs_def)
apply auto
apply  (drule_tac x = "pa" in spec)
apply  (force simp add: fun_upd_apply split add: split_if_asm)+
done
lemma noleak_all_add1msg_tail [simp]: "\<lbrakk>p \<noteq>  NA; p \<noteq>  NB\<rbrakk> \<Longrightarrow>  
noleak_all nB (b(p := b p @ [m]), s) = (noleak_all nB (b, s) \<and> noleak nB m)"
apply (simp add: noleak_all_def N_outputs_def)
apply auto
apply  (drule_tac x = "pa" in spec)
apply  (force simp add: fun_upd_apply split add: split_if_asm)+
done

(*>*)
lemma Bob_trusts_NSL_lemma: "\<lbrakk>Alice \<notin> bad; cs \<in> NSL_Runs\<rbrakk> \<Longrightarrow> 
       (\<forall>c\<in>set cs. constrain nB (snd c)) \<longrightarrow> (\<forall>c\<in>set cs. noleak_all nB c)"
(*<*)
apply (unfold NSL_Runs_def, erule runs_closed_comp_all_imp_all_induct,
      simp_all)
apply    (rule closed_System)
apply   (rule wf_isms_System)
apply  (fold K_Nil_def o_def)
apply  (simp add: System_def System_ism_defs del: o_apply o_def)
apply  (force simp add: noleak_all_def (* init_noleak *))
apply (erule_tac V = "?x : ?A" in thin_rl)
apply (erule_tac V = "Ball (set cs) (noleak_all ?nB)" in thin_rl)
apply (erule_tac V = "Ball (set cs) (constrain nB \<circ> snd)" in thin_rl)
apply (drule in_SystemD)
apply (safe del: conjI)
apply (simp_all add: System_ism_defs)
apply (erule_tac [!] System_trans_elims)
apply (safe del: conjI)
apply simp_all

apply (subgoal_tac "nA \<noteq> nB")
apply  (clarsimp simp add: noleak_all_def N_outputs_def)
apply (clarsimp simp add: constrain_def Alice_Wait_nA_def)

apply (erule_tac [!] V = "constrain ?nB ?s" in thin_rl)

apply (clarify, erule disjE)
apply  (clarsimp simp add: constrain_def Alice_Fine_with_def)
apply  (erule_tac x = "AnA sa" in allE)
apply  simp
apply (clarsimp simp add: noleak_all_def)

apply (clarsimp simp add: constrain_def Bob_Resp_only_to_def)
apply (clarsimp simp add: noleak_all_def N_outputs_def 
                          fun_upd_apply split add: split_if_asm)

apply (clarsimp simp add: noleak_all_def)

apply (simp add: noleak_all_def)
apply (clarify)
apply (erule_tac x = "ch" in allE)
apply (force simp add: I_inputs_def N_outputs_def)

apply (simp add: noleak_all_def)
apply (clarsimp)
apply (erule_tac x = "p" in allE)
apply (simp add: fun_upd_apply split add: split_if_asm)
apply (erule (2) synth_analz_maintains_noleak)

apply (clarsimp simp add: noleak_all_def fun_upd_apply split add: split_if_asm)
done
(*>*)

lemma Bob_trusts_NSL_lemma': "\<lbrakk>Alice \<notin> bad; cs \<in> NSL_Runs\<rbrakk> \<Longrightarrow> 
 (\<forall>(b,s)\<in>set cs. Bob_Resp_only_to Alice s \<and> \<not>Alice_Fine_with Bob nB s \<and> 
                 \<not>Alice_Wait_nA nB s) \<longrightarrow> 
 (\<forall>(b,s)\<in>set cs. \<forall>m\<in>(\<Union>set`(b`{AI,IA,IB,BI}))\<union>Intruder_known s. noleak nB m)"
(*<*)
apply (drule (1) Bob_trusts_NSL_lemma)
apply (clarsimp simp add: constrain_def noleak_all_def N_outputs_def)
apply (erule impE)
apply  (clarsimp, fast)
apply (erule ballE) back
prefer 2 apply blast
apply simp
apply blast
done
(*>*)

theorem Bob_trusts_NSL: "\<lbrakk>Alice \<notin> bad; (b,s)#cs \<in> NSL_Runs\<rbrakk> \<Longrightarrow> 
  (\<exists>nA. Bob_state   s  = (Conn, \<lparr>Bpeer = Alice, BnA = nA, BnB = nB\<rparr>)) \<longrightarrow>
  (\<exists>(b',s')\<in>set cs. 
  (\<exists>nA. Alice_state s' = (Fine, \<lparr>Apeer = Bob  , AnA = nA, AnB = nB\<rparr>)))"
(*<*)
apply (erule NSL_Runs_System_past_induct)
apply (auto elim!: System_trans_elims)
apply (drule (1) Bob_trusts_NSL_lemma)
apply safe
prefer 2 (* by contradiction: *)
apply  (clarsimp simp add: noleak_all_def)
apply  (erule_tac x = "IB" in allE)
apply  (simp add: N_outputs_def (* leak_Crypt_Bob_nB *))
apply  fast
apply (simp add: constrain_def)
apply (rule conjI)
apply  (drule_tac c = "(a,ba)" in Alice_Wait_nA_not_BnB, simp, simp)
apply  (simp)
apply (simp add: Alice_Fine_with_def Bob_Resp_only_to_def)
apply (erule disjE)
apply  simp
apply (drule (1) bspec)
apply  simp 
apply (force dest: Bob_state_Resp_eq)
done
(*>*)

end
