theory Test imports "ism/ISM_package"
begin

use_thy "../awe-0.4/Extensions/AWE"
use_thy "../noninfluence/Noninfluence2Dom"

record "state" =
  HighData :: nat
  LowData :: nat
datatype "msg" = INPUTHIGH nat | INPUTLOW nat | OUTPUTHIGH | OUTPUTLOW | OUT nat

datatype "domain" = H | L
datatype "action" = A1 | A2 | A3 | A4
datatype "io" = HighIn | HighOut | LowIn | LowOut

ism Trivial =
  ports "io"
    inputs  "{HighIn,LowIn}"
    outputs "{HighOut,LowOut}"
  messages "msg"
  labels "action"
  domains "domain"
  states
    data "state"

  transitions
  ih:
    label A1
    dom H
    in HighIn "[INPUTHIGH n]"
    post HighData := n
  il:
    label A2
    dom L
    in HighIn "[INPUTLOW n]"
    post LowData := n
  oh:
    label A3
    dom H
    pre "HighData s > 5"
    in  HighIn  "[OUTPUTHIGH]"
    out HighOut "[OUT (HighData s)]"
  ol:
    label A4
    dom L
    pre "LowData s > 5"
    in  HighIn  "[OUTPUTLOW]"
    out HighOut "[OUT (LowData s)]"
  ol2:
    label A4
    dom L
    pre "LowData s = 0"
    in  HighIn  "[OUTPUTLOW]"
    out HighOut "[OUT 17]"
  ol3:
    label A4
    dom L
    pre "LowData s > 15"
    in  HighIn  "[OUTPUTLOW]"
    out HighOut "[OUT ((LowData s)+15)]"

thm "dom_def"
thm "transs.intros"
consts policy     :: "domain => domain => bool" ("(_\ \<leadsto>\ _)"(*<*) [46,46] 45(*>*))
primrec
"(L \<leadsto> y) = True"
"(H \<leadsto> y) = (y = H)"

consts uwr :: "state => domain => state => bool" ("(_\ \<sim>_\<sim>\ _)"(*<*) [46,46,46] 45(*>*))
primrec
"(s1 \<sim>H\<sim> s2) = (s1=s2)"
"(s1 \<sim>L\<sim> s2) = ((LowData s1) = (LowData s2))"

theorem uwr_refl: "s \<sim>u\<sim> s"
by(induct u, auto)

theorem policy_refl: "u \<leadsto> u"
by(induct u, auto)

constdefs S0 :: "state set"
         "S0 \<equiv> init ism"
constdefs reachable :: "state set"
         "reachable \<equiv> reach ism"

lemma uwr_s0: "\<forall>s0\<in>S0. s0 \<sim>u\<sim> s0"
by(simp add:uwr_refl)

lemma S0_nonempty: "\<exists>s. s \<in> S0"
by (simp add: S0_def ism_def ism.make_def)

lemma S0_reachable: "s0 \<in> S0 \<Longrightarrow> s0 \<in> reachable"
by (force simp add: S0_def reachable_def reach_def dest: runs.init)

constdefs Step :: "action \<Rightarrow> (state \<times> state) set"
         "Step a \<equiv> ISM.Step transs a"

constdefs "Output" :: "action => domain => state => ((io, msg) msgss) set"
          "Output a u s \<equiv> {out. \<exists> out' \<in> ISM.Outputs transs a s. out= (case u of H \<Rightarrow> \<lambda>x. if x=HighOut then (out' x) else [] | L \<Rightarrow> \<lambda>x. if x=LowOut then (out' x) else [])}"
constdefs empty_output :: "(io, msg) msgss"
         "empty_output \<equiv> \<currency>"

lemma trans_eq: "transs = trans ism"
by (auto simp add: ism_def ism.make_def)

theorem Step_preserves_reachability: "\<lbrakk>s \<in> reachable; (s, s') \<in> Step a\<rbrakk> \<Longrightarrow> s' \<in> reachable"
by (auto simp add: ISM.Step_preserves_reachability reachable_def Step_def trans_eq)

thymorph s: Noninfluence2Dom \<longrightarrow> Test
  maps [("System.state" \<mapsto> "state"), ("System.action" \<mapsto> "action"),
        ("System.output" \<mapsto> "(io, msg) msgss"), ("System.domain" \<mapsto> "domain")]
       [("Generics2Dom.L" \<mapsto> "Test.domain.L"), ("Generics2Dom.H" \<mapsto> "Test.domain.H"),
        ("Generics.reachable" \<mapsto> "Test.reachable"), ("SystemNoninterference.S0" \<mapsto> "Test.S0")]

t_instantiate "Noninfluence2Dom" mapping s

translate_thm "Noninfluence2Dom.Two_domains_sym.Low_Step_consistent_Local_respect_left_implies_Noninfluence" as "ni" along s'

(*
Falls der Automat funktional ist kann man auch die folgenden Theoreme verwenden:
translate_thm "Noninterference.simple_Noninterference" as "" along s'
translate_thm "Generics.Step_functional.uni_wStep_consistent" as "" along s'
translate_thm "Generics.Step_functional.uni_wStep_respect" as "" along s'
*)

lemma out_ineq_simp: "(\<lambda>x. if x = LowOut then out x else []) \<noteq> (\<lambda>x. if x = LowOut then out' x else []) \<Longrightarrow> out LowOut \<noteq> out' LowOut"
by (auto simp add: expand_fun_eq, case_tac x, auto)

lemma out_empty_simp: "out LowOut = [] \<Longrightarrow> (\<lambda>x. if x = LowOut then out x else []) = \<currency>"
by (simp add: ext)

lemma out_notempty_simp: "\<currency> \<noteq> (\<lambda>x. if x = LowOut then out x else []) \<Longrightarrow> out LowOut \<noteq> []"
by (auto simp add: expand_fun_eq, case_tac x, auto)

lemma oc': "\<lbrakk>s \<in> reachable; t \<in> reachable; LowData s = LowData t; x \<in> wOutput a L s\<rbrakk> \<Longrightarrow> x \<in> wOutput a L t"
apply (auto simp add: wOutput_def empty_output_def Output_def Outputs_def)
apply   (drule spec, auto)
apply   (drule out_ineq_simp)
apply   (force elim: transs.elims)
apply  (drule spec, auto)
apply  (rule out_empty_simp)
apply  (force elim: transs.elims)
apply (drule spec, auto)
apply (drule out_notempty_simp)
apply (force elim: transs.elims)
done

lemma oc: "output_consistent"
apply (clarsimp simp add:output_consistent_def empty_output_def)
apply (case_tac u, clarsimp simp add: dom_def)
apply  (case_tac a, clarsimp+)
apply (auto dest: oc')
apply  (auto simp add: wOutput_def Output_def empty_output_def Outputs_def)
apply  (rule out_empty_simp)
apply  (force elim: transs.elims)
apply (drule spec, auto)
apply (drule out_notempty_simp)
apply (force elim: transs.elims)
done

lemma td: "Two_domains_sym"
apply (unfold Two_domains_sym_def uwr_refl_def Two_domains_axioms_def uwr_sym_def, auto)
apply   (case_tac u, auto)
apply  (case_tac x, auto)
apply (case_tac u, auto)
done

lemma lsc: "Low_Step_consistent"
apply (clarsimp simp add: Low_Step_consistent_def Step_def ISM.Step_def)
apply (erule transs.cases, auto simp add: dom_def)
apply    (force intro: transs.intros)+
done

lemma lrl: "Local_respect_left"
apply (clarsimp simp add: Local_respect_left_def Step_def ISM.Step_def)
apply (case_tac a, auto simp add: dom_def)
apply  (auto elim: transs.elims)
apply (case_tac u, auto elim: transs.elims)
done

theorem "Noninfluence"
by (rule ni, auto simp add: td oc lsc lrl)

end
